// Copyright (c) 2019 WSO2 Inc. (http://www.wso2.org) All Rights Reserved.
//
// WSO2 Inc. licenses this file to you under the Apache License,
// Version 2.0 (the "License"); you may not use this file except
// in compliance with the License.
// You may obtain a copy of the License at
//
// http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing,
// software distributed under the License is distributed on an
// "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
// KIND, either express or implied.  See the License for the
// specific language governing permissions and limitations
// under the License.

import ballerina/internal;
import ballerina/io;

public type FuncBodyParser object {
    BirChannelReader reader;
    TypeParser typeParser;
    map<VariableDcl> localVarMap;
    map<VariableDcl> globalVarMap;
    TypeDef?[] typeDefs;

    public function __init(BirChannelReader reader,  TypeParser typeParser, map<VariableDcl> globalVarMap, map<VariableDcl> localVarMap, TypeDef?[] typeDefs) {
        self.reader = reader;
        self.typeParser = typeParser;
        self.localVarMap = localVarMap;
        self.globalVarMap = globalVarMap;
        self.typeDefs = typeDefs;
    }

    public function parseBB() returns BasicBlock {
        var id = self.reader.readStringCpRef();
        var numInstruction = self.reader.readInt32() - 1;
        Instruction?[] instructions = [];
        int i = 0;
        while (i < numInstruction) {
            instructions[i] = self.parseInstruction();
            i += 1;
        }

        return { id: { value: id }, instructions: instructions, terminator: self.parseTerminator() };
    }

    public function parseEE() returns ErrorEntry {
        return { trapBB: self.parseBBRef(), errorOp: self.parseVarRef() };
    }

    public function parseInstruction() returns Instruction {
        DiagnosticPos pos = parseDiagnosticPos(self.reader);
        var kindTag = self.reader.readInt8();
        InstructionKind kind = INS_KIND_CONST_LOAD;
        // this is hacky to init to a fake val, but ballerina dosn't support un intialized vars
        if (kindTag == INS_ARRAY_STORE) {
            kind = INS_KIND_ARRAY_STORE;
            var lhsOp = self.parseVarRef();
            var keyOp = self.parseVarRef();
            var rhsOp = self.parseVarRef();
            FieldAccess mapStore = {pos:pos, kind:kind, lhsOp:lhsOp, keyOp:keyOp, rhsOp:rhsOp};
            return mapStore;
        } else if (kindTag == INS_MAP_STORE) {
            kind = INS_KIND_MAP_STORE;
            var lhsOp = self.parseVarRef();
            var keyOp = self.parseVarRef();
            var rhsOp = self.parseVarRef();
            FieldAccess mapStore = {pos:pos, kind:kind, lhsOp:lhsOp, keyOp:keyOp, rhsOp:rhsOp};
            return mapStore;
        } else if (kindTag == INS_MAP_LOAD) {
            kind = INS_KIND_MAP_LOAD;
            var lhsOp = self.parseVarRef();
            var keyOp = self.parseVarRef();
            var rhsOp = self.parseVarRef();
            FieldAccess mapLoad = {pos:pos, kind:kind, lhsOp:lhsOp, keyOp:keyOp, rhsOp:rhsOp};
            return mapLoad;
        } else if (kindTag == INS_OBJECT_STORE) {
            kind = INS_KIND_OBJECT_STORE;
            var lhsOp = self.parseVarRef();
            var keyOp = self.parseVarRef();
            var rhsOp = self.parseVarRef();
            FieldAccess objectStore = {pos:pos, kind:kind, lhsOp:lhsOp, keyOp:keyOp, rhsOp:rhsOp};
            return objectStore;
        } else if (kindTag == INS_OBJECT_LOAD) {
            kind = INS_KIND_OBJECT_LOAD;
            var lhsOp = self.parseVarRef();
            var keyOp = self.parseVarRef();
            var rhsOp = self.parseVarRef();
            FieldAccess objectLoad = {pos:pos, kind:kind, lhsOp:lhsOp, keyOp:keyOp, rhsOp:rhsOp};
            return objectLoad;
        } else if (kindTag == INS_ARRAY_LOAD) {
            kind = INS_KIND_ARRAY_LOAD;
            var lhsOp = self.parseVarRef();
            var keyOp = self.parseVarRef();
            var rhsOp = self.parseVarRef();
            FieldAccess arrayLoad = {pos:pos, kind:kind, lhsOp:lhsOp, keyOp:keyOp, rhsOp:rhsOp};
            return arrayLoad;
        } else if (kindTag == INS_NEW_ARRAY) {
            var bType = self.typeParser.parseType();
            kind = INS_KIND_NEW_ARRAY;
            var lhsOp = self.parseVarRef();
            var sizeOp = self.parseVarRef();
            NewArray newArray = {pos:pos, kind:kind, lhsOp:lhsOp, sizeOp:sizeOp, typeValue:bType};
            return newArray;
        } else if (kindTag == INS_NEW_MAP) {
            var bType = self.typeParser.parseType();
            kind = INS_KIND_NEW_MAP;
            var lhsOp = self.parseVarRef();
            NewMap newMap = {pos:pos, kind:kind, lhsOp:lhsOp, typeValue:bType};
            return newMap;
        } else if (kindTag == INS_NEW_TABLE) {
            var bType = self.typeParser.parseType();
            kind = INS_KIND_NEW_TABLE;
            var lhsOp = self.parseVarRef();
            var columnsOp = self.parseVarRef();
            var dataOp = self.parseVarRef();
            var indexColOp = self.parseVarRef();
            var keyColOp = self.parseVarRef();
            NewTable newTable = {pos:pos, kind:kind, lhsOp:lhsOp, columnsOp: columnsOp, dataOp: dataOp, indexColOp:
            indexColOp, keyColOp: keyColOp, typeValue:bType};
            return newTable;
        } else if (kindTag == INS_NEW_INST) {
            var defIndex = self.reader.readInt32();
            kind = INS_KIND_NEW_INST;
            var lhsOp = self.parseVarRef();
            NewInstance newInst = {pos:pos, kind:kind, lhsOp:lhsOp, typeDef: self.findTypeDef(defIndex)};
            return newInst;
        } else if (kindTag == INS_TYPE_CAST) {
            kind = INS_KIND_TYPE_CAST;
            var lhsOp = self.parseVarRef();
            var rhsOp = self.parseVarRef();
            TypeCast typeCast = {pos:pos, kind:kind, lhsOp:lhsOp, rhsOp:rhsOp};
            return typeCast;
        } else if (kindTag == INS_IS_LIKE) {
            kind = INS_KIND_IS_LIKE;
            var bType = self.typeParser.parseType();
            var lhsOp = self.parseVarRef();
            var rhsOp = self.parseVarRef();
            IsLike isLike = {pos:pos, kind:kind, typeValue:bType, lhsOp:lhsOp, rhsOp:rhsOp};
            return isLike;
        } else if (kindTag == INS_TYPE_TEST) {
            kind = INS_KIND_TYPE_TEST;
            var bType = self.typeParser.parseType();
            var lhsOp = self.parseVarRef();
            var rhsOp = self.parseVarRef();
            TypeTest typeTest = {pos:pos, kind:kind, typeValue:bType, lhsOp:lhsOp, rhsOp:rhsOp};
            return typeTest;
        } else if (kindTag == INS_CONST_LOAD){
            //TODO: remove redundent
            var bType = self.typeParser.parseType();
            kind = INS_KIND_CONST_LOAD;
            var lhsOp = self.parseVarRef();

            int | string | boolean | float value = 0;
            if (bType is BTypeInt || bType is BTypeByte) {
                value = self.reader.readIntCpRef();
            } else if (bType is BTypeString) {
                value = self.reader.readStringCpRef();
            } else if (bType is BTypeBoolean) {
                value = self.reader.readBoolean();
            } else if (bType is BTypeFloat) {
                value = self.reader.readFloatCpRef();
            }
            ConstantLoad constLoad = {pos:pos, kind:kind, lhsOp:lhsOp, typeValue:bType, value:value};
            return constLoad;
        } else if (kindTag == INS_MOVE){
            kind = INS_KIND_MOVE;
            var rhsOp = self.parseVarRef();
            var lhsOp = self.parseVarRef();
            Move move = {pos:pos, kind:kind, lhsOp:lhsOp, rhsOp:rhsOp};
            return move;
        } else if (kindTag == INS_NEW_ERROR) {
            kind = INS_KIND_NEW_ERROR;
            var lhsOp = self.parseVarRef();
            var reasonOp = self.parseVarRef();
            var detailsOp = self.parseVarRef();
            NewError newError = {pos:pos, kind:kind, lhsOp:lhsOp, reasonOp:reasonOp, detailsOp:detailsOp};
            return newError;
        } else if (kindTag == INS_NEW_XML_ELEMENT) {
            kind = INS_KIND_NEW_XML_ELEMENT;
            var lhsOp = self.parseVarRef();
            var startTagOp = self.parseVarRef();
            var endTagOp = self.parseVarRef();
            var defaultNsURIOp = self.parseVarRef();
            NewXMLElement newXMLElement = {pos:pos, kind:kind, lhsOp:lhsOp, startTagOp:startTagOp, endTagOp:endTagOp, 
                                           defaultNsURIOp:defaultNsURIOp};
            return newXMLElement;
        } else if (kindTag == INS_NEW_XML_TEXT) {
            kind = INS_KIND_NEW_XML_TEXT;
            var lhsOp = self.parseVarRef();
            var textOp = self.parseVarRef();
            NewXMLText newXMLText = {pos:pos, kind:kind, lhsOp:lhsOp, textOp:textOp};
            return newXMLText;
        } else if (kindTag == INS_NEW_XML_COMMENT) {
            kind = INS_KIND_NEW_XML_COMMENT;
            var lhsOp = self.parseVarRef();
            var textOp = self.parseVarRef();
            NewXMLComment newXMLComment = {pos:pos, kind:kind, lhsOp:lhsOp, textOp:textOp};
            return newXMLComment;
        } else if (kindTag == INS_NEW_XML_PI) {
            kind = INS_KIND_NEW_XML_PI;
            var lhsOp = self.parseVarRef();
            var dataOp = self.parseVarRef();
            var targetOp = self.parseVarRef();
            NewXMLPI newXMLPI = {pos:pos, kind:kind, lhsOp:lhsOp, dataOp:dataOp, targetOp:targetOp};
            return newXMLPI;
        } else if (kindTag == INS_NEW_XML_QNAME) {
            kind = INS_KIND_NEW_XML_QNAME;
            var lhsOp = self.parseVarRef();
            var localnameOp = self.parseVarRef();
            var nsURIOp = self.parseVarRef();
            var prefixOp = self.parseVarRef();
            NewXMLQName newXMLQName = {pos:pos, kind:kind, lhsOp:lhsOp, localnameOp:localnameOp, nsURIOp:nsURIOp, 
                                       prefixOp:prefixOp};
            return newXMLQName;
        } else if (kindTag == INS_NEW_STRING_XML_QNAME) {
            kind = INS_KIND_NEW_STRING_XML_QNAME;
            var lhsOp = self.parseVarRef();
            var stringQNameOp = self.parseVarRef();
            NewStringXMLQName newStringXMLQName = {pos:pos, kind:kind, lhsOp:lhsOp, stringQNameOp:stringQNameOp};
            return newStringXMLQName;
        } else if (kindTag == INS_XML_SEQ_STORE) {
            kind = INS_KIND_XML_SEQ_STORE;
            var lhsOp = self.parseVarRef();
            var rhsOp = self.parseVarRef();
            XMLAccess xmlSeqStore = {pos:pos, kind:kind, lhsOp:lhsOp, rhsOp:rhsOp};
            return xmlSeqStore;
        } else if (kindTag == INS_XML_SEQ_LOAD) {
            kind = INS_KIND_XML_SEQ_LOAD;
            var lhsOp = self.parseVarRef();
            var keyOp = self.parseVarRef();
            var rhsOp = self.parseVarRef();
            FieldAccess xmlLoad = {pos:pos, kind:kind, lhsOp:lhsOp, keyOp:keyOp, rhsOp:rhsOp};
            return xmlLoad;
        } else if (kindTag == INS_XML_LOAD) {
            kind = INS_KIND_XML_LOAD;
            var lhsOp = self.parseVarRef();
            var keyOp = self.parseVarRef();
            var rhsOp = self.parseVarRef();
            FieldAccess xmlLoad = {pos:pos, kind:kind, lhsOp:lhsOp, keyOp:keyOp, rhsOp:rhsOp};
            return xmlLoad;
        } else if (kindTag == INS_XML_LOAD_ALL) {
            kind = INS_KIND_XML_LOAD_ALL;
            var lhsOp = self.parseVarRef();
            var rhsOp = self.parseVarRef();
            XMLAccess xmlLoadAll = {pos:pos, kind:kind, lhsOp:lhsOp, rhsOp:rhsOp};
            return xmlLoadAll;
        } else if (kindTag == INS_XML_ATTRIBUTE_STORE) {
            kind = INS_KIND_XML_ATTRIBUTE_STORE;
            var lhsOp = self.parseVarRef();
            var keyOp = self.parseVarRef();
            var rhsOp = self.parseVarRef();
            FieldAccess xmlAttrStore = {pos:pos, kind:kind, lhsOp:lhsOp, keyOp:keyOp, rhsOp:rhsOp};
            return xmlAttrStore;
        } else if (kindTag == INS_XML_ATTRIBUTE_LOAD) {
            kind = INS_KIND_XML_ATTRIBUTE_LOAD;
            var lhsOp = self.parseVarRef();
            var keyOp = self.parseVarRef();
            var rhsOp = self.parseVarRef();
            FieldAccess xmlAttrLoad = {pos:pos, kind:kind, lhsOp:lhsOp, keyOp:keyOp, rhsOp:rhsOp};
            return xmlAttrLoad;
        } else if (kindTag == INS_FP_LOAD) {
            kind = INS_KIND_FP_LOAD;
            var lhsOp = self.parseVarRef();
            var pkgId = self.reader.readModuleIDCpRef();
            var name = self.reader.readStringCpRef();

            var mapCount = self.reader.readInt32();
            VarRef?[] maps = [];
            int j = 0;
            while (j < mapCount) {
                maps[j] = self.parseVarRef();
                j += 1;
            }

            VariableDcl?[] params = [];
            var numVars = self.reader.readInt32();
            int i = 0;
            while (i < numVars) {
                var dcl = parseVariableDcl(self.reader, self.typeParser);
                params[i] = dcl;
                i += 1;
            }
            FPLoad fpLoad = {pos:pos, kind:kind, lhsOp:lhsOp, pkgID:pkgId, name:{ value: name }, params:params, closureMaps:maps};
            return fpLoad;
        } else if (kindTag == INS_TYPEOF) {
            kind = INS_KIND_TYPEOF;
            var rhsOp = self.parseVarRef();
            var lhsOp = self.parseVarRef();
            UnaryOp typeofNode = {pos:pos, kind:kind, lhsOp:lhsOp, rhsOp:rhsOp};
            return typeofNode;
        } else if (kindTag == INS_NOT) {
            kind = INS_KIND_NOT;
            var rhsOp = self.parseVarRef();
            var lhsOp = self.parseVarRef();
            UnaryOp typeofNode = {pos:pos, kind:kind, lhsOp:lhsOp, rhsOp:rhsOp};
            return typeofNode;
        } else {
            return self.parseBinaryOpInstruction(kindTag, pos);
        }
    }

    public function parseTerminator() returns Terminator {
        DiagnosticPos pos = parseDiagnosticPos(self.reader);
        var kindTag = self.reader.readInt8();
        if (kindTag == INS_BRANCH){
            TerminatorKind kind = TERMINATOR_BRANCH;
            var op = self.parseVarRef();
            BasicBlock trueBB = self.parseBBRef();
            BasicBlock falseBB = self.parseBBRef();
            Branch branch = {pos:pos, falseBB:falseBB, kind:kind, op:op, trueBB:trueBB};
            return branch;
        } else if (kindTag == INS_GOTO){
            TerminatorKind kind = TERMINATOR_GOTO;
            GOTO goto = {pos:pos, kind:kind, targetBB:self.parseBBRef()};
            return goto;
        } else if (kindTag == INS_RETURN){
            TerminatorKind kind = TERMINATOR_RETURN;
            Return ret = {pos:pos, kind:kind};
            return ret;
        } else if (kindTag == INS_CALL){
            TerminatorKind kind = TERMINATOR_CALL;
            var isVirtual = self.reader.readBoolean();
            var pkgId = self.reader.readModuleIDCpRef();
            var name = self.reader.readStringCpRef();
            var argsCount = self.reader.readInt32();
            VarRef?[] args = [];
            int i = 0;
            while (i < argsCount) {
                args[i] = self.parseVarRef();
                i += 1;
            }
            var hasLhs = self.reader.readBoolean();
            VarRef? lhsOp = ();
            if (hasLhs){
                lhsOp = self.parseVarRef();
            }

            BasicBlock thenBB = self.parseBBRef();
            Call call = {pos:pos, args:args, kind:kind, isVirtual: isVirtual,
                         lhsOp:lhsOp, pkgID:pkgId,
                         name:{ value: name }, thenBB:thenBB};
            return call;
        } else if (kindTag == INS_ASYNC_CALL){
            TerminatorKind kind = TERMINATOR_ASYNC_CALL;
            var pkgId = self.reader.readModuleIDCpRef();
            var name = self.reader.readStringCpRef();
            var argsCount = self.reader.readInt32();
            VarRef?[] args = [];
            int i = 0;
            while (i < argsCount) {
                args[i] = self.parseVarRef();
                i += 1;
            }
            var hasLhs = self.reader.readBoolean();
            VarRef? lhsOp = ();
            if (hasLhs){
                lhsOp = self.parseVarRef();
            }

            BasicBlock thenBB = self.parseBBRef();
            AsyncCall call = {pos:pos, args:args, kind:kind, lhsOp:lhsOp, pkgID:pkgId, 
                                name:{ value: name }, thenBB:thenBB};
            return call;
        } else if (kindTag == INS_PANIC) {
            TerminatorKind kind = TERMINATOR_PANIC;
            var errorOp = self.parseVarRef();
            Panic panicStmt = { pos:pos, kind:kind, errorOp:errorOp };
            return panicStmt;
        } else if (kindTag == INS_WAIT) {
            TerminatorKind kind = TERMINATOR_WAIT;
            var exprCount = self.reader.readInt32();
            VarRef?[] exprs = [];
            int i = 0;
            while (i < exprCount) {
                exprs[i] = self.parseVarRef();
                i += 1;
            }
            VarRef lhsOp = self.parseVarRef();
            Wait waitIns = {pos:pos, exprList:exprs, kind:kind, lhsOp:lhsOp};
            return waitIns;
        } else if (kindTag == INS_FP_CALL) {
            TerminatorKind kind = TERMINATOR_FP_CALL;
            VarRef fp = self.parseVarRef();
            
            var argsCount = self.reader.readInt32();
            VarRef?[] args = [];
            int i = 0;
            while (i < argsCount) {
                args[i] = self.parseVarRef();
                i += 1;
            }

            var hasLhs = self.reader.readBoolean();
            VarRef? lhsOp = ();
            if (hasLhs){
                lhsOp = self.parseVarRef();
            }

            BasicBlock thenBB = self.parseBBRef();
            FPCall fpCall = {pos:pos, kind:kind, fp:fp, lhsOp:lhsOp, args:args, thenBB:thenBB};
            return fpCall;
        }
        error err = error("term instrucion kind " + kindTag + " not impl.");
        panic err;
    }

    public function parseVarRef() returns VarRef {
        VarKind kind = parseVarKind(self.reader);
        VarScope varScope = parseVarScope(self.reader);
        string varName = self.reader.readStringCpRef();

        var decl = getDecl(self.globalVarMap, self.localVarMap, varScope, varName, kind);
        return {typeValue : decl.typeValue, variableDcl : decl};
    }

    public function parseBBRef() returns BasicBlock {
        return { id: { value: self.reader.readStringCpRef() } };
    }

    public function parseBinaryOpInstruction(int kindTag, DiagnosticPos pos) returns BinaryOp {
        BinaryOpInstructionKind kind = BINARY_ADD;
        if (kindTag == INS_ADD){
            kind = BINARY_ADD;
        } else if (kindTag == INS_SUB){
            kind = BINARY_SUB;
        } else if (kindTag == INS_MUL){
            kind = BINARY_MUL;
        } else if (kindTag == INS_DIV){
            kind = BINARY_DIV;
        } else if (kindTag == INS_EQUAL){
            kind = BINARY_EQUAL;
        } else if (kindTag == INS_NOT_EQUAL){
            kind = BINARY_NOT_EQUAL;
        } else if (kindTag == INS_GREATER_THAN){
            kind = BINARY_GREATER_THAN;
        } else if (kindTag == INS_GREATER_EQUAL){
            kind = BINARY_GREATER_EQUAL;
        } else if (kindTag == INS_LESS_THAN){
            kind = BINARY_LESS_THAN;
        } else if (kindTag == INS_LESS_EQUAL){
            kind = BINARY_LESS_EQUAL;
        } else if (kindTag == INS_AND){
            kind = BINARY_AND;
        } else if (kindTag == INS_OR){
            kind = BINARY_OR;
        } else {
            error err = error("instrucion kind " + kindTag + " not impl.");
            panic err;
        }

        var rhsOp1 = self.parseVarRef();
        var rhsOp2 = self.parseVarRef();
        var lhsOp = self.parseVarRef();
        BinaryOp binaryOp = {pos:pos, kind:kind, lhsOp:lhsOp, rhsOp1:rhsOp1, rhsOp2:rhsOp2};
        return binaryOp;
    }

    function findTypeDef(int defIndex) returns TypeDef {
        var typeDef = self.typeDefs[defIndex];
        if(typeDef is TypeDef){
            return typeDef;
        } else {
            error err = error("can't find type def for index : " + defIndex);
            panic err;
        }
    }

};

function getDecl(map<VariableDcl> globalVarMap, map<VariableDcl> localVarMap, VarScope varScope, string varName, 
        VarKind kind) returns VariableDcl {
    if (varScope == VAR_SCOPE_GLOBAL) {
        var possibleDcl = globalVarMap[varName];
        if (possibleDcl is VariableDcl) {
            return possibleDcl;
        } else {
            error err = error("global var missing " + varName);
            panic err;
        }
    }

    // for self referrence, create a dummy varDecl
    if (kind == VAR_KIND_SELF) {
        VariableDcl varDecl = { kind : kind, 
                                varScope : varScope, 
                                name : {value : varName}
                              };
        return varDecl;
    }

    var possibleDcl = localVarMap[varName];
    if (possibleDcl is VariableDcl) {
        return possibleDcl;
    } else {
        error err = error("local var missing " + varName);
        panic err;
    }
}

public function parseVariableDcl(BirChannelReader reader, TypeParser typeParser) returns VariableDcl {
    VarKind kind = parseVarKind(reader);
    var typeValue = typeParser.parseType();
    var name = reader.readStringCpRef();
    VariableDcl dcl = {
        typeValue: typeValue,
        name: { value: name },
            kind: kind
        };
    return dcl;
}
